"""
Main entry point for the command line.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from fusion_platform.base import Base


def fusion_platform():
    """
    Main entry point for command line.
    """
    Base.main()
