"""
Main entry point for package.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from fusion_platform.base import Base

# Main entry point when the package is executed.
if __name__ == "__main__":
    Base.main()
